#ifndef CAD_MATH_H
#define CAD_MATH_H

#include "HybLib/OSysDep.h"
#ifndef llabs
#define llabs(a) (a>0?a:-a)
#endif
#ifndef RoundH
// #define RoundH(dd) (int)((fabs(dd-(int)(dd)))>=0.5?((int)(dd)+((dd<0)?-1:1)):((int)(dd)))
#define RoundH(dd) (long)((fabs(dd-(long)(dd)))>=0.5?((long)(dd)+((dd<0)?-1:1)):((long)(dd)))
#define RoundHP(dd) (long)((dd-(long)(dd))>=0.5?(long)dd+1:(long)dd)
#endif

#ifndef RoundI64
#define RoundI64(dd) (INT64)((fabs(dd-(INT64)(dd)))>=0.5?((INT64)(dd)+((dd<0)?-1:1)):((INT64)(dd)))
#endif


#ifndef TO_RAD
#define TO_RAD(x)		(M_PI*(x)/180.0)
#endif

#ifndef TO_DEG
#define TO_DEG(x)		(180.0*(x)/M_PI)
#endif

#ifndef TO_RADIANS
#define TO_RADIANS(x)	TO_RAD(x)
#endif

#ifndef TO_DEGREES
#define TO_DEGREES(x)	TO_DEG(x)
#endif

#define M_PI_360 6.283185307179586
#define M_PI_270 4.71238898038469
#define M_PI_90  1.57079632679489661923


// ƽת
#define AREASIZE_SQM_TO_HECTARE(x)	(x / 10000.0)
// ƽתĶ
#define AREASIZE_SQM_TO_MU(x)	(x * 15.0 / 10000.0)
#define DOUBLE_EQ_DIFF 0.00000001
#define PT_INVALID_XY -100000000

class CadMath
{
public:
	CadMath(void);
	~CadMath(void);
};

class RVector
{
public:
	double x,y,z;
public:
	inline RVector() :
		x(0.0), y(0.0), z(0.0){
	}
    /**
     * Constructor for a point with given coordinates.
     */
    RVector(double vx, double vy, double vz = 0.0);

	void setPolar(double radius, double angle);
	RVector operator +(const RVector& v) const;
	RVector operator -(const RVector& v) const;
	RVector operator *(double s) const;
	RVector operator /(double s) const;
	void operator +=(const RVector& v);
	void operator -=(const RVector& v);
	void operator *=(double s);
	void operator /=(double s);
    RVector getNegated() const;

	RVector operator -() const;
	bool operator ==(const RVector& v) const;
	bool operator !=(const RVector& v) const {
		return !operator==(v);
	}

	double getDistanceTo(const RVector& v) const;
	void setAngle(double a);
	double getAngle() const;
	double getAngleTo(const RVector& v) const;
	double getAngle3dTo(const RVector& v) const;
    double getMagnitude() const;
    double getSquaredMagnitude() const;
    double getMagnitude2d() const;
    static double getDotProduct(const RVector& v1, const RVector& v2);
	static RVector getCrossProduct(const RVector& v1, const RVector& v2);
	double Dot(RVector& v);
	RVector Cross(RVector& v);

	RVector rotate(double rotation);
	RVector rotate(double rotation, const RVector& center);
	RVector rotate3d(double rotation, RVector rvLineStart, RVector rvLineEnd);
	RVector rotate3d(double rotation, RVector rvDir);
	RVector stretch3d(RVector rvLineStart, RVector rvLineEnd);
	double GetRotateDirAndAngle(RVector rvTo, RVector& rvDir);
	double GetRotateDirAndAngle(RVector rv0, RVector rvTo, RVector& rvDir);
	RVector getLerp(const RVector& v, double t) const;
	double getDistanceToLine(const RVector& v1, const RVector& v2) const;
	double getDistanceToLine1(const RVector& v1, double dAngle);
	RVector getMiddleVt(const RVector& v1, const RVector& v2);
	RVector scale(double factor, const RVector& center);
	RVector scaleF(double factor, const RVector& center);
	double GetScale(RVector& rvCenter, RVector& rvOther);
	RVector GetLatLngAlt(double dRadius);
	void MakeEarthXyz(double lat, double lng, double dRadius);

	bool GetCrossToBall(RVector rvTo, double dBallRadius, RVector& rvNear);
};

class RArc
{
public:
    RArc();
    RArc(const RVector& center, double radius, double startAngle,
            double endAngle, bool reversed = false);
	static RArc createFrom2PBulge(const RVector& startPoint,
		const RVector& endPoint,
		double bulge);
	static RArc createFrom3Points(const RVector& rvStart, const RVector& rv, const RVector& rvEnd);
	virtual RVector getStartPoint() const;
	virtual RVector getEndPoint() const;
	RVector getPoint(double dAngle) const;
	double getSweep() const;
	double getBulge() const;
public:
	/**
	* \getter{getCenter}
	* \setter{setCenter}
	*/
	RVector center;
	/**
	* \getter{getRadius}
	* \setter{setRadius}
	*/
	double radius;
	/**
	* \getter{getStartAngle}
	* \setter{setStartAngle}
	*/
	double startAngle;
	/**
	* \getter{getEndAngle}
	* \setter{setEndAngle}
	*/
	double endAngle;
	/**
	* \getter{isReversed}
	* \setter{setReversed}
	*/
	bool reversed;
};


// Բ
double GetEllipseLength(double a, double b);
double GetEllipseSimLength(double a, double b);
double GetEllipseArcLength(double a, double b, double dStartangle, double dEndangle, double dLenCalc = -1, double dLenSim = -1);
double GetEllipseArcAngle(double a, double b, double dStartangle, double dLength, double dLenCalc = -1, double dLenSim = -1);

CRect64 Rect32To64(RECT &rect);
CRect Rect64To32(RECTH64 &rect);

// double Get1PiRangeRad(double fRad);
// double Get1PiRangeDegree(double fDegree);
// double Get2PiRangeDegree(double fDegree);
bool GetRatioPoit2(double x1, double y1, double x2, double y2, double x3, double &y3);
bool GetRatioPoit3(double x1, double y1, double z1, double x2, double y2, double z2, double x3, double &y3, double &z3);
bool CsysTransformRestoration2f(double dx1, double dy1, double dScale, double dRotateDeg, double &dx2, double &dy2);
void CircleRotationDegFormula(double &dx, double &dy, double dRotateDeg);
void CircleRotationRadFormula(double &dx, double &dy, double dRotateRad);
void GetSquareMeanValue(double bl, double br, double tl, double tr, double fRatioX, double fRatioY, double &fMean);

double StandrandAngle(double dAngle, double dBase = 0, bool bUpBase = false);
double StandrandRadian(double dRad, double dBase = 0, bool bUpBase = false);
double CadAngleTrans(double dAngle);
void RectAddPoint64(RECTH64& re, INT64 x, INT64 y);
void RectAddPoint(RECT& re,int x,int y);
void RectAddSize64(RECTH64& re, INT64 cx, INT64 cy);
void RectAddSize(RECT& re, int cx, int cy);
RVector GetXyInTiltEllipse(RVector rvCenter, RVector rvEnd, double dAxisScale, double dAngle);
void GetXyInTiltEllipse(RVector rvCenter, RVector rvEnd, double dAxisScale, double dAngle, double& x, double &y);
void GetXyInEllipse64(CRect64 reArea, double dAngle, double& x, double &y);
void GetXyInEllipse(CRect reArea, double dAngle, double& x, double &y);
double GetEllipseParamAngle64(CRect64 reArea, double dAngle);
void GetXyInEllipseParam64(CRect64 reArea, double dAngle, double& x, double &y);

double GetLineAngle(double x1, double y1, double x2, double y2);
double GetLineDist(double x1,double y1,double x2,double y2);
double GetLineDistI64(INT64 x1, INT64 y1, INT64 x2, INT64 y2);
double GetPointToLineDist(CPoint pt1, CPoint pt2, CPoint pt3, CPoint *pPtIntersec = NULL);
double GetPoint64ToLineDist(CPoint64 pt1, CPoint64 pt2, CPoint64 pt3, CPoint64 *pPtIntersec = NULL);
int PtToArcWidth(CRect64 reArea, INT64 x, INT64 y, INT64 iArcX1, INT64 iArcY1, INT64 iArcX2, INT64 iArcY2, double dBulge);
int PtToArcWidth(CRect64 reArea, INT64 x, INT64 y, double dStartAngle, double dEndAngle);
CRect64 GetArcDrawArea64(CRect64 reArc, double dStartAngle, double dEndAngle);
CRect GetArcDrawArea(CRect reArc,double dStartAngle,double dEndAngle);
void GetCadBulgeAngle(double xCenter,double yCenter,double x1,double y1,double x2,double y2,double dBulge,double& dStartAngle,double& dEndAngle);
double GetCadBulgeFromMapPoint(double cx, double cy, double x1, double y1, double x2, double y2, bool bClockwise);
double GetPointDis(POINT u1, POINT v1);
bool IsLineParallel(POINT u1, POINT u2, POINT v1, POINT v2);
bool TwoLineInteraction(POINT point1, POINT point2, POINT point3, POINT point4, POINT& pointInterSection);
bool TwoLineInteraction64(POINTH64 point1, POINTH64 point2, POINTH64 point3, POINTH64 point4, POINTH64& pointInterSection);
bool TwoLineInteractionRv(RVector point1, RVector point2, RVector point3, RVector point4, RVector& pointInterSection);
bool GetLineRectInteraction(RVector pt1, RVector pt2, CRect re, RVector* pRvInt, int* pnRvInt);
bool GetLineRectInteraction64(RVector pt1, RVector pt2, CRect64 re, RVector* pRvInt, int* pnRvInt);
RVector GetPointToLineInteraction(RVector pt1, RVector pt2, RVector pt3);
void GetBoxRotateVector(RVector* prvCenter, double iWidth, double iHeight, double fAngle, RVector* prvLt, RVector* prvRt, RVector* prvLb, RVector* prvRb, RVector* prvLeft, RVector* prvRight, RVector* prvTop, RVector* prvBottom, RVector* prvRotate, int* piRadius);
CRect GetRectRotateLb(CRect re, double fAngle, RVector* prv4);
CRect64 GetRectRotateLb64(CRect64 re, double fAngle, RVector* prv4);
CRect GetRectRotate(CRect re, RVector rvCenter, double fAngle, RVector* prv4);
CRect64 GetRectRotate64(CRect64 re, RVector rvCenter, double fAngle, RVector* prv4);
void RVectorRotate(RVector rvCenter, RVector& rv, double fAngle);
void GetPlaneEdushiOffset(POINTH64 ptSrc1, POINTH64 ptSrc2, POINTH64 ptSrc3, POINTH64 ptSrc4, POINTH64 ptDst1, POINTH64 ptDst2, POINTH64 ptDst3, POINTH64 &ptDst4);
bool IsSameSlope(POINT pt1, POINT pt2, POINT pt3);

CRect64 GetEllipseDrawBox(RVector rvCenter, RVector rvEnd, double dAxisScale, int iWidth, double dStartAngle, double dEndAngle);
bool PtInEllipse(RVector rvCenter, RVector rvEnd, double dAxisScale, double x, double y);
bool PtInEllipse(RVector rvCenter, RVector rvEnd, double dAxisScale, double x, double y, double dStartAngle, double dEndAngle);
bool PtInEllipse(CRect reArea, int x, int y);
bool Pt64InEllipse(CRect64 reArea, INT64 x, INT64 y);
bool PtInSector(RVector rvCenter, RVector rvEnd, double dAxisScale, double x, double y, double dStartAngle, double dEndAngle);
bool PtInSector(CRect reArea, int x, int y, double dStartAngle, double dEndAngle);
bool Pt64InSector(CRect64 reArea, INT64 x, INT64 y, double dStartAngle, double dEndAngle);

int GetLineInteractionCircle(RVector rv1, RVector rv2, RVector rvCenter, double dRadius, RVector& rvInt1, RVector& rvInt2);
int TwoCircleInteraction(RVector rv1, double dRadius1, RVector rv2, double dRadius2, RVector& rvInt1, RVector& rvInt2);

bool IntersectTriangle(RVector rvOrg, RVector rvDir, RVector rv0, RVector rv1, RVector rv2, bool bPlaneCheckOnly, double *t, double *u, double *v);
bool IntersectTriangle(RVector rvStart, RVector rvEnd, RVector rv0, RVector rv1, RVector rv2, RVector& rvInt);
bool IntersectPlane(RVector rvStart, RVector rvEnd, RVector rv0, RVector rv1, RVector rv2, RVector& rvInt);
bool Is2DTriangleClockwise(RVector rv0, RVector rv1, RVector rv2);
double Get2DTriangleArea(RVector& rv0, RVector& rv1, RVector& rv2);
double Get2DQuadArea(RVector& rvLt, RVector& rvRt, RVector& rvLb, RVector& rvRb);
double Get2DQuadArea(RVector& rvLt, RVector& rvRt, RVector& rvLb, RVector& rvRb, RVector& rvM);
bool Is2DQuadIntersectBox(RVector& rvLt, RVector& rvRt, RVector& rvLb, RVector& rvRb, double dxLt, double dyLt, double dxRb, double dyRb);

double GetTriangleAngleByEdgeSize(double dAngleEdge, double dEdge1, double dEdge2);

RVector GetPointScaleRvInBox(RVector rvCenter, RVector rvX, RVector rvY, RVector rvPoint);
RVector GetPointInBoxByScaleRv(RVector rvCenter, RVector rvX, RVector rvY, RVector rvScale);


// ȡ, ٻȡôߵĴ
bool GetVerticalPedal(double x1, double y1, double x2, double y2, double xx, double yy, double &px, double &py);
// ֪Բ
bool GetCircelCenterByPoint3(double x1, double y1, double x2, double y2, double x3, double y3, double &cx, double &cy);
double GetCycleRangeNum(double fNum, double fMin, double fMax, double dCycle);
double GetDegDiff(double d1, double d2);

void PtAddRvBox(RVector& rvLt, RVector& rvRb, double x, double y);

#endif
